@@ -82,6 +82,9 @@ static XOP nv_push_xop;
 static XOP nv_pop_xop;
 static XOP nv_peek_xop;
 
+/* Method-to-op optimization */
+static Perl_check_t orig_entersub_check;
+
 /* ============================================
    Magic vtables
    ============================================ */
@@ -1023,6 +1026,126 @@ static OP* pp_heap_func_size(pTHX) {
     RETURN;
 }
 
+/* ============================================
+   Method-to-op optimization via PL_check hook
+   ============================================ */
+
+/* Forward declarations */
+static OP* pp_heap_push(pTHX);
+static OP* pp_heap_pop(pTHX);
+static OP* pp_heap_peek(pTHX);
+static OP* pp_heap_size(pTHX);
+
+/*
+ * Intercept OP_ENTERSUB at compile time.
+ * If it's a method_named call for push/pop/peek/size on a heap,
+ * replace with custom op.
+ *
+ * Tree structure for $h->push(5):
+ *   entersub-+-pushmark
+ *            |-padsv[$h]        <- heap
+ *            |-const[IV 5]      <- value (for 2-arg methods)
+ *            `-method_named[PV "push"]
+ */
+static OP* heap_entersub_check(pTHX_ OP *entersubop) {
+    OP *pushop, *firstarg, *methop, *prevop;
+    const char *meth_name;
+    STRLEN meth_len;
+    OP *heapop, *valop = NULL;
+    OP *newop;
+    OP *(*ppfunc)(pTHX) = NULL;
+    bool is_2arg = FALSE;
+    int nargs = 0;
+
+    /* Only for ENTERSUB ops */
+    if (entersubop->op_type != OP_ENTERSUB)
+        return orig_entersub_check(aTHX_ entersubop);
+
+    /* Navigate to first child - might be wrapped in ex-list */
+    pushop = cUNOPx(entersubop)->op_first;
+    if (!OpHAS_SIBLING(pushop))
+        pushop = cUNOPx(pushop)->op_first;
+
+    /* pushop should be pushmark, first sibling is first argument */
+    if (pushop->op_type != OP_PUSHMARK)
+        return orig_entersub_check(aTHX_ entersubop);
+
+    firstarg = OpSIBLING(pushop);
+    if (!firstarg)
+        return orig_entersub_check(aTHX_ entersubop);
+
+    /* Walk to find method_named - it's the last sibling */
+    prevop = pushop;
+    methop = firstarg;
+    while (OpHAS_SIBLING(methop)) {
+        prevop = methop;
+        methop = OpSIBLING(methop);
+        nargs++;
+    }
+
+    /* Must end with method_named */
+    if (methop->op_type != OP_METHOD_NAMED)
+        return orig_entersub_check(aTHX_ entersubop);
+
+    /* Get method name */
+#if PERL_VERSION_GE(5,22,0)
+    meth_name = SvPV_const(cMETHOPx_meth(methop), meth_len);
+#else
+    meth_name = SvPV_const(cSVOPx(methop)->op_sv, meth_len);
+#endif
+
+    /* Check if it's one of our methods */
+    if (meth_len == 4 && strEQ(meth_name, "push")) {
+        ppfunc = pp_heap_push;
+        is_2arg = TRUE;
+    } else if (meth_len == 3 && strEQ(meth_name, "pop")) {
+        ppfunc = pp_heap_pop;
+        is_2arg = FALSE;
+    } else if (meth_len == 4 && strEQ(meth_name, "peek")) {
+        ppfunc = pp_heap_peek;
+        is_2arg = FALSE;
+    } else if (meth_len == 4 && strEQ(meth_name, "size")) {
+        ppfunc = pp_heap_size;
+        is_2arg = FALSE;
+    } else {
+        return orig_entersub_check(aTHX_ entersubop);
+    }
+
+    /* Verify argument count: 1 arg for 1-arg methods, 2 for 2-arg */
+    if (is_2arg && nargs != 2)
+        return orig_entersub_check(aTHX_ entersubop);
+    if (!is_2arg && nargs != 1)
+        return orig_entersub_check(aTHX_ entersubop);
+
+    /* Extract ops:
+     * For 1-arg: firstarg is heap
+     * For 2-arg: firstarg is heap, OpSIBLING(firstarg) is value
+     */
+    heapop = firstarg;
+
+    if (is_2arg) {
+        valop = OpSIBLING(heapop);
+
+        /* Unlink heap and val from chain */
+        OpMORESIB_set(pushop, methop);
+        OpLASTSIB_set(heapop, NULL);
+        OpLASTSIB_set(valop, NULL);
+
+        newop = newBINOP(OP_CUSTOM, 0, heapop, valop);
+        newop->op_ppaddr = ppfunc;
+    } else {
+        /* Unlink heap from chain */
+        OpMORESIB_set(pushop, methop);
+        OpLASTSIB_set(heapop, NULL);
+
+        newop = newUNOP(OP_CUSTOM, 0, heapop);
+        newop->op_ppaddr = ppfunc;
+    }
+
+    op_free(entersubop);
+    return newop;
+}
+
 /* Call checkers */
 typedef OP* (*heap_ppfunc)(pTHX);
 
@@ -1514,6 +1637,10 @@ XS_EXTERNAL(boot_heap) {
         newXS("heap::make_heap_max", XS_make_heap_max, __FILE__);
     }
 
+    /* Install PL_check hook for method-to-op optimization */
+    orig_entersub_check = PL_check[OP_ENTERSUB];
+    PL_check[OP_ENTERSUB] = heap_entersub_check;
+
 #if PERL_VERSION_GE(5,22,0)
     Perl_xs_boot_epilog(aTHX_ ax);
 #else
